function convertTypedArray(sourceArray, targetConstructor) {
    var buffer = new ArrayBuffer(sourceArray.byteLength);
    new sourceArray.constructor(buffer).set(sourceArray);
    return new targetConstructor(buffer);
}

function G711Encoder(inputBuffer) {
    var MAX_16BIT = 32767;
    var outputBuffer = new ArrayBuffer(inputBuffer.byteLength / 2);
    var outputView = new DataView(outputBuffer);
    for (var i = 0; i < outputView.byteLength; i++) {
        let bitMask = 0xF;
        let signBit = 0x0;
        let exponent = 7;
        let mantissa = 0;
        let sample = inputBuffer.getInt16(i * 2, true);
        sample /= -2;
        if (sample < 0) {
            sample = -sample;
            signBit = 0x80;
        }
        if (sample > MAX_16BIT) {
            sample = MAX_16BIT;
        }
        for (; (sample & bitMask) != bitMask && exponent >= 5; bitMask >>= 1, exponent--);
        mantissa = (sample >> (exponent == 4 ? 1 : exponent - 4)) & 0xF;
        outputView.setUint8(i, (signBit | ((exponent - 5) << 4) | mantissa) ^ 0x55);
    }
    return outputBuffer;
}

var audioData = {
    resample: function(channelData, inputSampleRate, outputSampleRate, outputSampleBits) {
        var decompressedData = this.decompress(channelData, channelData.length, inputSampleRate, outputSampleRate);
        var length = decompressedData.length * (outputSampleBits / 8);
        var buffer = new ArrayBuffer(length);
        var outputView = new DataView(buffer);
        this.reshapeData(outputSampleBits, 0, decompressedData, outputView);
        return outputView;
    },
    decompress: function(inputData, length, inputSampleRate, outputSampleRate) {
        var decompressed = new Float32Array(length);
        decompressed.set(inputData, 0);
        var sampleRatio = parseInt(inputSampleRate / outputSampleRate);
        var resampledLength = decompressed.length / sampleRatio;
        var resampledData = new Float32Array(resampledLength);
        var index = 0;
        var inputIndex = 0;
        while (index < resampledLength) {
            resampledData[index] = decompressed[inputIndex];
            inputIndex += sampleRatio;
            index++;
        }
        return resampledData;
    },
    reshapeData: function(outputSampleBits, startIndex, inputData, outputView) {
        var sample = 0;
        for (var i = 0; i < inputData.length; i++, startIndex += outputSampleBits / 8) {
            sample = Math.max(-1, Math.min(1, inputData[i]));
            outputView.setInt16(startIndex, sample < 0 ? sample * 32768 : sample * 32767, true);
        }
        return outputView;
    }
};

function Recorder(options) {
    this.context = null;
    this.stream = null;
    this.audioInput = null;
    this.analyser = null;
    this.recorder = null;
    this.destination = null;
    this.inputSampleRate = 48000;
    this.inputSampleBits = 16;
    this.outputSampleRate = 8000;
    this.outputSampleBits = 16;
    this.config = options;
}

Recorder.prototype.setNewOption = function(newConfig) {
    this.outputSampleRate = this.config.sampleRate;
    this.outputSampleBits = this.config.sampleBits;
}

Recorder.prototype.startRecord = function(callback) {
    var self = this;
    this.context = new (window.AudioContext || window.webkitAudioContext)();
    let createProcessor = this.context.createScriptProcessor || this.context.createJavaScriptNode;
    this.inputSampleRate = this.context.sampleRate;
    this.recorder = createProcessor.apply(this.context, [4096, 1, 1]);
    this.recorder.onaudioprocess = function(event) {
        var resampledData = audioData.resample(event.inputBuffer.getChannelData(0), self.inputSampleRate, self.outputSampleRate, 16);
        callback(G711Encoder(resampledData));
    };
    navigator.mediaDevices.getUserMedia({
        audio: true
    }).then(stream => {
        this.audioInput = this.context.createMediaStreamSource(stream);
        this.stream = stream;
    }).then(() => {
        this.audioInput.connect(this.recorder);
        this.destination = this.context.createMediaStreamDestination();
        this.recorder.connect(this.destination);
    }).catch(error => {
        console.log('error');
    });
}

Recorder.prototype.stopRecord = function() {
    if (this.context) {
        this.context.close();
        this.context = null;
    }
}

function WASMTalker(options) {
    this.url = null;
    this.recorder = null;
    this.websocket = null;
    this.seqNum = 0;
}

WASMTalker.prototype.checkEnv = function() {
    var currentUrl = window.location.href;
    if (currentUrl.indexOf('https') == -1 && currentUrl.indexOf('127.0.0.1') == -1 && currentUrl.indexOf('localhost') == -1)
        return '由于浏览器安全策略，非https或者localhost访问，对讲不可用';
    else
        return '';
}

WASMTalker.prototype.talkbackStart = function(url) {
    this.url = url;
    var self = this;
    var currentSeqNum = self.seqNum;
    if (self.recorder)
        self.recorder.stopRecord();
    if (self.websocket)
        self.websocket.close();
    if (url.indexOf('ws://') < 0 && url.indexOf('wss://') < 0)
        return;
    var ws = self.websocket = new WebSocket(url);
    ws.binaryType = 'arraybuffer';
    ws.onopen = function(event) {
        console.log(event);
    };
    ws.onclose = function(event) {
        console.log(event);
    };
    ws.onmessage = function(message) {
        if (currentSeqNum != self.seqNum) {
            ws.close();
            return;
        }
    };
    this.recorder = new Recorder();
    this.recorder.startRecord(function(data) {
        if (self.websocket.readyState == 1)
            self.websocket.send(data);
    });
    return;
}

WASMTalker.prototype.talkbackStop = function() {
    if (this.recorder) {
        this.recorder.stopRecord();
        this.recorder = null;
    }
    if (this.websocket) {
        this.websocket.close();
        this.websocket = null;
    }
    return;
}
